import java.util.*;


/**
 * calculation is a help class to calculate different positions
 * @author Annika Niehage
 * @version $Id: Calculation.java,v 1.3 2002/07/14 17:15:17 Niehage Exp $
 **/
abstract public class Calculation
{

    protected final double KICKABLEDISTANCE = 0.7;
    protected final CVector GOALPOSITION = new CVector(53, 0);
    protected final double DECAYPLAYER = 0.4;
    protected final double DECAYBALL = 0.94;

    protected Action action;
    protected Worldmodell data;
    protected int playerNum;
    protected int  absAngle;
    protected int neckAngle;
    protected int bodyAngle;
    protected CVector ownPosition;
    protected CVector standPosition;
    protected CVector startPosition;
    protected PVector ballPosition;
    protected CVector absBallPosition;  
    protected CVector ballSpeed;
    protected CVector ballFuturePosition;
    protected CVector ballMeetingPoint;
    protected CVector ownSpeed;
    protected int timeCycle;
    protected PVector oldBallPosition;
    protected boolean seeingBall;
    protected double dashPower;   



    /**
     * the constructor initalizes a Calculation object
     **/
    public Calculation( CommunicationChannel com, Worldmodell data)
    {
	action = new Action(com);
	this.data = data;
	seeingBall = false;
	oldBallPosition = new PVector();
	timeCycle = 0;
	ballMeetingPoint = new CVector();

    }


    /**
     * calculates own position, angle and ball data
     **/
    public void calcAll()
    {
	data.calcAll();
	ownPosition = data.getOwnPosition();
	absAngle = data.getDirection();
	neckAngle = data.getNeckAngle();
	bodyAngle = data.getBodyAngle();
	
	if (ballPosition!=null)
	    oldBallPosition = ballPosition;
	ballPosition = data.getBallPosition();
	
	data.newSeeMsg = false;

	if (ballPosition != null) {
	    seeingBall = true;
	    absBallPosition = ownPosition.add(ballPosition.toCVector(absAngle));
	    calcBallFuturePosition();
	    calcBallMeetingPoint();
	}
	else
	    seeingBall = false;
    }


    /**
     * calculates where the ball will probably be in the next cycle
     **/
    protected void calcBallFuturePosition()
    {
	data.calcBallSpeed();
	ballSpeed = data.getBallSpeed();
	
	absBallPosition = ownPosition.add(ballPosition.toCVector(absAngle));
	
	if(ballSpeed.length() == 0)
	    ballFuturePosition = absBallPosition;

	else
	    ballFuturePosition = new CVector(absBallPosition.x + ballSpeed.x * DECAYBALL, 
						    absBallPosition.y + ballSpeed.y * DECAYBALL);
    }



    /**
     * calculates where another player will be one cycle later
     *
     * @param PlayerListElement player: the object which future position I would like to know
     * @return CVector objFuturePosition: absolute future position of the object
     **/
    protected CVector calcObjFuturePosition(PlayerListElement player)
    {
	CVector objSpeed = data.calcObjSpeed(player);
	CVector absPlayerPosition = ownPosition.add(player.position.toCVector(absAngle));
	CVector objFuturePosition = new CVector(absPlayerPosition.x + objSpeed.x * DECAYPLAYER,
						absPlayerPosition.y + objSpeed.y * DECAYPLAYER);
	
	return objFuturePosition;
    }



    /**
     * calculates where the ball can be met best
     **/
    protected void calcBallMeetingPoint()
    {


	if (ballPosition.distance > 10)
	{
	    
	    ballMeetingPoint.x = absBallPosition.x + ballSpeed.x * 14.062406564;   // (1- Math.pow(DECAYBALL, 30))/(1-DECAYBALL)
	    ballMeetingPoint.y = absBallPosition.y + ballSpeed.y * 14.062406564;   // (1- Math.pow(DECAYBALL, 30))/(1-DECAYBALL)

	    
	}
	else
	{
	    if (ballPosition.distance > 4)
		ballMeetingPoint = ballFuturePosition;
	    else
		ballMeetingPoint = absBallPosition;
	}

    }
    


    /**
     * how much to accelerate and turns the player in the right direction
     * @param CVector position: position where I want to dash to
     **/
    protected void calcDashPowerAndTurn(CVector position)
    {
	if (position.sub(ownPosition).length() < 0.5)
	    return;

	ownSpeed = data.getOwnSpeed();
	double distance = position.sub(ownPosition).length();
	double speed = ownSpeed.length();
	
	//turns the player in the direction of given position
	int turnAngle = calcAngleToPos(position);

	if(distance > 1)
	    dashPower = 100;
	else
	    dashPower = 100/5*distance;	    

	if ( turnAngle>=7 || turnAngle<=-7 ) 
	{
	    action.turn(turnAngle);
	    absAngle = calcAngleModulo(absAngle + turnAngle);
	    bodyAngle = calcAngleModulo(bodyAngle + turnAngle);
	}
	else
	{
	    action.dash(dashPower);
	    ownPosition = ownPosition.add((new PVector(dashPower/100 , 0)).toCVector(absAngle));
	}
    }


    /**
     * calculates the angle to a given position
     *
     * @param CVector position: position to which I want to know the angle
     * @return int direction = AngleToPosition
     **/
    protected int calcAngleToPos(CVector position)
    {
	CVector diff = position.sub(ownPosition);


	if (diff.x == 0)
	{
	    if (diff.y == 0)
		return 0;
	    else
		if (diff.y < 0)
		    return -90;
		else
		    return 90;
	}
	else
	{
	    int direction;
	
	    direction = (int)(Math.atan((diff.y)/(diff.x))/Math.PI*180);
	    if (diff.x < 0.) 
		if (direction > 0.) 
		    direction = -180 + direction;
		else 
		    direction = 180 + direction;
	    
	    direction = calcAngleModulo(direction - absAngle);
	    
	    return direction;
	}
    }



    /**
     * provides that the given angle is between -180 and 180 degrees
     *
     **/
    protected int calcAngleModulo( int direction )
    {
	if (direction>180) 
	    direction = direction - 360;
	if (direction<-180) 
	    direction = direction + 360;
	return direction;
    }







    /**
     * decides where and how to kick
     **/
    abstract protected void calcKick();



    /**
     * gives his position to each player
     **/
    abstract protected void calcStandPosition();


} // class







