import java.io.*;
import java.net.*;


/**
 * The class handles UDP network traffic from/to the rcssserver
 * (formerly named <code>Communication</code>).
 *
 * @author Gerald Kuehne, Markus Beier
 * @author adapted to our client by Matthias Lang and Tim Oliver Wagner
 * @version $Id: CommunicationChannel.java,v 1.3 2002/07/11 23:26:27 Niehage Exp $
 **/
public class CommunicationChannel
{
    // socket for incoming, outgoing messages
    private DatagramSocket socket;
    // IP address of the server
    InetAddress serverAddress;
    // port of the server
    private int serverPort;

    /**
     * Creates a new <code>CommunicationChannel</code> instance.
     *
     * @param serverName The address of the server specified by its name or its IP address
     * @param serverPort The port of the server
     **/
    public CommunicationChannel(String serverName, int serverPort) {
        try {
            // construct datagram socket and
            // bind it to any available port on the local machine
            this.socket        = new DatagramSocket();
            // set target address
            this.serverAddress = InetAddress.getByName(serverName);
            // set target port
            this.serverPort    = serverPort;
        } catch (SocketException se) {
            System.err.println("Communication: Could not create datagram socket.");
            System.exit(1);
        } catch (UnknownHostException uhe) {
            System.err.println("Communication: Could not determine the server's IP address.");
            System.exit(1);
        }
    }


    /**
     * Sends a <code>String</code> to the server
     *
     * @param message a <code>String</code> value containing the message which
     * has to be sent to the server.
     **/
    public void send(String message) {
        try {
            // convert the message to a byte array
            byte[] buffer = message.getBytes();
            // create the packet from the buffer and set the server's IP address and port
            DatagramPacket packet =
                new DatagramPacket(buffer, buffer.length, serverAddress, serverPort);
            // send the packet
            socket.send(packet);
        }
	catch (IOException ioe) {
            System.err.println("Communication: Could not send datagram packet - IOException occurs.");
        }
	catch (SecurityException se) {
            System.err.println("Communication: Could not send datagram packet - SecurityException occurs.");
	}
	catch (Exception e) {
            System.err.println("Communication: Could not send datagram packet - unexpected Exception occurs.");
	}
    }


    /**
     * Receives a message from the server and stores it in the message queue.
     **/
    public String receive() {
        // allocate buffer for server message
        byte[] buffer = new byte[1500];
        // create datagram packet for receiving
        DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
        try {
            // receive packet, the method blocks until a datagram arrives
            socket.receive(packet);
        }
	catch (IOException ioe) {
            System.err.println("Commmunication: Could not receive datagram packet.");
        }
	catch (Exception e) {
            System.err.println("Communication: Could not send datagram packet - unexpected Exception occurs.");
	}


        // get the server port for the following actions
        serverPort = packet.getPort();
        // get and return the message
	return (new String(packet.getData()) );
    }
}
