import java.util.*;

/**
 * A class which search the relevant information and put it in the Worldmodell class
 * The constuctor of the Worldmodell class parse the parts into the right datatype
 *
 * @author Tim Oliver Wagner
 * @version $Id: DivideString.java,v 1.3 2002/07/11 23:26:27 Niehage Exp $
 **/

public class DivideString {

    private int currentPosition, time, timeBody;
    private String input;
    private String teamname;
    private Worldmodell wm;
    private String[] nearestFlag = new String[5];
    private String[] stringArrayBuffer;
    private String stringBuffer;
    private double distNearestFlag = 100000, bufferFlagDist;
    private boolean alreadySetLine = false, alreadySetBallPos = false;
    private Strategy strategy;
    /**
     * A constructor for a DivideString class object.
     *
     * @param teamname contains the team name of the player who uses the class to divide the input strings
     * @param wm       contains the reference to the worldmodel object which should save the current see, sense_body, ...
     *                 information
     **/
    DivideString (String teamname, Worldmodell wm, Strategy strategy) {
	this.teamname = teamname;
	this.wm = wm;
	this.strategy = strategy;
    }


    /**
     * This method recognize the kind of the current string and starts the corresponding process
     *
     * @param input contains the reference to the current string element
     **/
    public void runDivideString (String input) {
//	try {
	    this.input = input;
	    String kindOfString = input.substring(1, currentPosition = input.indexOf(' ', 1));

	    if (kindOfString.equals("sense_body")) divideSENSE_BODYString();
	    else 
	    if (kindOfString.equals("see")) divideSEEString();
	    else if (kindOfString.equals("hear")) divideAndSendHEARString();
	    else if (kindOfString.equals("init")) divideINITString();
/*	}
	catch (Exception e) {
	    WriteErrorMsg.writeErrMsg(input, e, "writeErrorMsg.DivideString.java");
	    System.err.println(e);
	}
*/

    }


    /**
     * divideSEEString divide the see information in suggestive parts.
     **/
    private void divideSEEString()
    {

	StringTokenizer inputST = new StringTokenizer(input, "()"), helpST;
	int counter;

	try {
	    time = Integer.parseInt(inputST.nextToken().substring(4).trim());
        } catch (Exception e) {
	    time = 0;
            System.out.println(e);
        }

	while (inputST.hasMoreTokens()) {
	    counter = 0;
	    stringArrayBuffer = new String[7];

	    stringArrayBuffer[counter++] = inputST.nextToken();
	    if (!inputST.hasMoreTokens())
		break;
	    helpST = new StringTokenizer(inputST.nextToken(), " ");

	    while (helpST.hasMoreTokens() && counter < 7) {
		stringArrayBuffer[counter++] = helpST.nextToken();
	    }
	    if (inputST.hasMoreTokens())
		inputST.nextToken();
	    handleSeePart(stringArrayBuffer);
	}

	    if (nearestFlag != null)
		wm.setFlag(nearestFlag);
	    if (!alreadySetBallPos)
		wm.setBallData(null);

	    wm.seeSomething = (alreadySetLine && !(distNearestFlag == 100000));
	    wm.setTimeAndPlayerLists(time);
	    wm.newSeeMsg = true;

	resetInstanceVariable();

	strategy.calcAll();
	strategy.act();
    }

    private void resetInstanceVariable () {
	distNearestFlag = 100000;
	alreadySetLine = false;
	alreadySetBallPos = false;
    }

    private void handleSeePart(String[] theCurrentString) {
	switch (theCurrentString[0].charAt(0)) {
	    case 'f': {
		if (theCurrentString[2] != null) {
		    bufferFlagDist = Float.parseFloat(theCurrentString[1]);
		    if (bufferFlagDist < distNearestFlag) {
			nearestFlag = theCurrentString;
			distNearestFlag = bufferFlagDist;
		    }
		}
		break;
	    }
	    case 'p': {
		if (theCurrentString[0].length() > 2)
		    transPlayer(theCurrentString);
		break;
	    }
	    case 'b': {
		wm.setBallData(theCurrentString);
		alreadySetBallPos = true;
		break;
	    }
	    case 'l': {
		if (!alreadySetLine) {
		    wm.setDirection(theCurrentString);
		    alreadySetLine = true;
		}
		break;
	    }
	}
    }

    /**
     * We know, that the object contains a player.
     * The method decides which team the current "player-object" belongs to and transfer it into the right list.
     **/
    private void transPlayer (String[] theCurrentString) {
	if (-1 == theCurrentString[0].indexOf(teamname, 2))
	    wm.putOpponent(theCurrentString);
	else
	    wm.putTeamMember(theCurrentString);
    }


    /**
     * Put the necessary information into the world model object
     * Only the information head_angle is necessary for our Strategy-Team
     **/
    private void divideSENSE_BODYString() {
	int head_angle;
	int begin = input.indexOf("head_angle", 67), end;

	begin = input.indexOf(' ', begin);
	end = input.indexOf(')', begin);

	head_angle = Integer.parseInt(input.substring(begin+1, end));

	wm.setNeckAngle(head_angle);

	StringTokenizer inputST = new StringTokenizer(input, "()"), helpST;
	int counter;

	try {
	    timeBody = Integer.parseInt(inputST.nextToken().substring(11).trim());
        } catch (Exception e) {
	    timeBody = 0;
            System.out.println(e);
        }

	strategy.setTime(timeBody);

    }



    /**
     * Devides the hear information into useful strings and put the necessary hear information to the world model object!
     **/
    private void divideAndSendHEARString() {
	String[] output = new String[3];
	int begin = 5, end = input.indexOf(' ', begin+1), absEnd = input.lastIndexOf(')');
	int i;

	for (i = 0; i < 2; ++i) {
	    output[i] = input.substring(begin+1, end);
	    begin = end;
	    end = input.indexOf(' ', begin+1);
	}
	output[i] = input.substring(begin+1, input.lastIndexOf(')'));

	// *********************************** Test Ausgabe auf StdOut *************************************************
//	System.out.println("Begin of the HEAR-String");
//	for (i = 0; i < output.length; ++i) {
//	    System.out.println("1.:***" + output[i] + "***");
//	}
	// *************************************************************************************************************

	wm.setHearMessage(output);
    }

    /**
     * The information we get direct from the string, without divide it
     **/
    private void divideINITString() {
	wm.setInitValues(input);
	strategy.init();
    }


    /**
     * Ascertained the first occurence of char c1 or c2 (in respect of the input string), start searching at pos beginPos
     * 
     * @param  c1       contains the first char to verify
     * @param  c2       contains the second char to verify
     * @param  beginPos contains the position the search starts
     * @return int      returns the position of the first occurence from one of both 
     **/
    private int searchFirstIndexOf (char c1, char c2, int beginPos) {
	int a = input.indexOf(c1, beginPos), b = input.indexOf(c2, beginPos);

	if (a != -1 && b !=- 1) {
	    if (a < b) return a;
	    else if (b < a) return b;
	    else if (a == b) return a;
	}
	else if (a == -1 && b > -1) return b;
	else if (b == -1 && a > -1) return a;
	return -1;
    }
}
