import javax.swing.*;
import java.awt.*;
import java.awt.geom.*;
import java.util.*;

/**
*This class can draw a soccerfield into a JPanel and show data from a Worldmodell-Object in it.
*
*@author Alexander Mller
*@author Olga Mirkina
**/
public class Draw extends JPanel
{
    private Worldmodell player;  //In here the data that has to be visualized is stored
    private int playernumber = 11;  //...and its number (11 for no player)
    
    // The size variables of the field  
    static double  N;              //Enlarges the field by a certain value
    static double dx = 52.5;       //x coordinate field center
    static double dy = 34;         //y coordinate field center
    static int ab=3;               //field border
    
    /**
     *The Constructor gets the value enlarge by which the field has to be
     *enlarged.
     *
     *By extending JPanel the constructor creates one. The JPanel is created 
     *in the size the main JFrame reserves for it getting its enlarge factor.
     *
     *@param double enlarge gives an enlarge factor
     **/ 
    public Draw(double enlarge)
	{
	    N = enlarge;
	    setMinimumSize(new Dimension(212,378));
	}

    /**
     *SetFactor is a methode to set the enlarge factor if it changes.
     *
     *@param double enlarge gives a new enlarge factor
     **/
    public void setFactor(double enlarge)
	{
	    N = enlarge;
	}

    /**
     *If a Player is chosen this Methode sets the Player that has to be drawn.
     *His number is also set.
     *
     *@param player An Object of the Dataclass that has to be Visualized.
     *@param i The integer with the Playernumber-1
     **/
    public void setPlayer(Worldmodell player , int i)
	{
            this.player = player;
	    playernumber = i;
	}
    
    /**
     *If no Player is chosen this Methode sets playernumber on 11 so that 
     *no viualsization takes place.
     *
     *@param i 11 for no Player
     **/
    public void setPlayer(int i)
	{
	    playernumber = i;
	}

    /**
     *This method calls the methodes that draws the objects of the field like goals, 
     *penalty areas and so on.
     *
     *Paint runs everytime something has to be redrawn. This may be either a change like resizing
     *or a change in data. For a redraw after a change in data in another methode repaint() is called.
     *
     *@param Graphics g is an Object that paint gets automatically when an repaint()-call 
     *takes place
     **/
    public void paint(Graphics g)
	{
	    Dimension d = getSize();
	    d.height = (int)N*68*2;	//Given size
	    d.width = (int)N*105*2;
	    Image screen = createImage(d.width,d.height);	 //new Image
	    Graphics grafik = screen.getGraphics();		 //new Graphic-Object
	    Graphics2D g2d = (Graphics2D)grafik;		 //convert for Graphics2D
            //Fills the Image with drawable Objects
	    fillField(g2d);
            //Draws Image screen, dx and dy (=ab) away from the left upper edge
            //"this" is a drawing-rule
	    g.drawImage(screen,ab,ab,this);
	}

    /**
     *This Methode calls the Methodes to draw the Field and to draw the Players.
     *
     *Visualize is only called when playernumber is smaller than 11(No visualization)
     *
     *@param Graphics2D g2d is the Graphics variable of paint converted to a Graphics2D-
     *Object that extends Graphics
     **/
    public void fillField(Graphics2D g2d)
	{
	    drawField(g2d);
	    if (playernumber<11)
		visualize(g2d);
	    else;
	}

    /**
     *This Methode fills the Field-Image.
     *
     *It contains the objects that are needed to draw the field.
     *
     *@param Graphics2D Feld is the Graphics variable of paint converted to a Graphics2D-
     *Object that extends Graphics
     **/
    private void drawField(Graphics2D Feld)
	{
	    double dw;          //Field length
	    double db;          //Field width
	    double a1;          //y coordinate top of the goalposts
	    double a2;          //y coordinate top of the penalty area
	    double a3;          //x coordinate right goalposts
	    double a4;          //x coordinate right penalty area
	    double a5;          //width of the penalty area
	    double a6;          //length of the penalty area
	    double a7;          //width of the goalpost area
	    double a8;          //length of the goalpost area
	    
	    
	    //Giving values to the size variables  
	    dx = N*(ab+52.5);
	    dy = N*(ab+34);
	    dw = N*105;
	    db = N*68;
	    a1 = dy - N*11;
	    a2 = dy - N*20.175;
	    a3 = dx + N*46;
	    a4 = dx + N*36;
	    a5 = N*16.5;
	    a6 = N*40.35;
	    a7 = N*6.5;
	    a8 = N*22;
         
      
	    //Creating the objects 
	    Rectangle2D r2d = new Rectangle2D.Double(N*ab,N*ab,dw,db);
	    Rectangle2D torl = new Rectangle2D.Double(N*ab,dy-N*7.01,N*0.5,N*14.02);
	    Rectangle2D torr = new Rectangle2D.Double(dx+dw/2-N*0.5,dy-N*7.01,N*0.5,N*14.02);
	    Rectangle2D penl = new Rectangle2D.Double(N*ab,a2,a5,a6);
	    Rectangle2D penr = new Rectangle2D.Double(a4,a2,a5,a6);
	    Rectangle2D postl = new Rectangle2D.Double(N*ab,a1,a7,a8);
	    Rectangle2D postr = new Rectangle2D.Double(a3,a1,a7,a8);
	    Line2D mittel = new Line2D.Double(dx,N*ab,dx,N*ab+db);
	    
	    /*
	      The first two parameters give the position of the top left corner of a restriction box, 
	      the next two - the side lengths, then come the start and the end points of the arc 
	      (from 0 to 360 to build a circle). The last 0 means that the circle is not marked with
	      any lines within it.
	    */        
	    Arc2D kreis = new Arc2D.Double(dx-N*9.15, dy-N*9.15, N*18.3, N*18.3, 0, 360, 0);
	    Arc2D center = new Arc2D.Double(dx-N*0.5, dy-N*0.5, N, N, 0, 360, 0);
	    Arc2D strafl = new Arc2D.Double(N*(ab+10.5), dy-N*0.5, N*0.5, N*0.5, 0, 360, 0);
	    Arc2D strafr = new Arc2D.Double(N*ab+dw-N*11.5, dy-N*0.5, N*0.5, N*0.5, 0, 360, 0);
	    Arc2D tkreisl = new Arc2D.Double(N*ab+N*1.85, dy-N*9.15, N*18.3, N*18.3, -52.5, 105, 0);
	    Arc2D tkreisr = new Arc2D.Double(N*ab+dw-N*20.15, dy-N*9.15, N*18.3, N*18.3, 127.5, 105, 0);
	    Arc2D corner1 = new Arc2D.Double(N*ab-N, N*ab-N, N*2, N*2, -90, 90, 0);
	    Arc2D corner2 = new Arc2D.Double(N*ab-N, N*ab+db-N, N*2, N*2, 0, 90, 0);
	    Arc2D corner3 = new Arc2D.Double(N*ab+dw-N, N*ab-N, N*2, N*2, 180, 90, 0);
	    Arc2D corner4 = new Arc2D.Double(N*ab+dw-N, N*ab+db-N, N*2, N*2, 90, 90, 0);

	    //Painting the objects
	    Feld.setPaint(Color.green);
	    Feld.fill(r2d);
	    Feld.setPaint(Color.black);
	    Feld.draw(r2d);
	    Feld.fill(torl);
	    Feld.fill(torr);
	    Feld.draw(penl);
	    Feld.draw(penr);
	    Feld.draw(postl);
	    Feld.draw(postr);
	    Feld.draw(mittel);
	    Feld.draw(kreis);
	    Feld.fill(center);
	    Feld.draw(strafl);
	    Feld.draw(strafr);
	    Feld.draw(tkreisl);
	    Feld.draw(tkreisr);
	    Feld.draw(corner1);
	    Feld.draw(corner2);
	    Feld.draw(corner3);
	    Feld.draw(corner4);
	}

    /**
     *This method draws the chosen player and the objects he sees.
     *
     *@param Graphics2D gp is the Graphics variable of paint converted to a Graphics2D-
     *Object that extends Graphics
     **/
    private void visualize(Graphics2D gp)
	{
	    //AlphaComposite insures that gp is drawn above Feld from DrawField.
	    //This could give a conflict because both Objects have the same priority.
	    //Here gp gets the higher priority.
	    AlphaComposite ac = AlphaComposite.getInstance(AlphaComposite.SRC);
	    gp.setComposite(ac);
	    
	    //The data of the player-reference is stored in lokal variables.
	    CVector p = player.getOwnPosition(); //the own position
	    PVector b = player.getBallPosition(); //the relative ball position
	    LinkedList tm = player.getTeamMembers(); //a list of the seen team members
	    LinkedList op = player.getOpponents(); //a list of the seen Opponents
	    
	    //Draws the player that is to be visualized
	    gp.setPaint(Color.red); //in red 
	    Arc2D spieler = new Arc2D.Double(dx+N*(p.x-1), dy+N*(p.y-1), 2*N, 2*N, 0, 360, 0);

	    //Draws a line to show the body angle
	    CVector r = new CVector(2*Math.cos((Math.PI/180)*player.getBodyAngle()), 2*Math.sin((Math.PI/180)*player.getBodyAngle())).add(p);
	    Line2D richtung = new Line2D.Double(dx+N*p.x, dy+N*p.y, dx+N*r.x, dy+N*r.y);
	    
	    //Makes the visualise-objects visible
	    if (p!=null)
	    {
		gp.fill(spieler);
		gp.draw(richtung);
	    }
	    
	    //Draw the ball
	    gp.setPaint(Color.orange); //in orange
	    if ( b!=null )
	    {
		CVector Cb = b.toCVector(player.getDirection()).add(p);//get AbsPosition
		Arc2D ball = new Arc2D.Double(dx+N*(Cb.x-1) , dy+N*(Cb.y-1), 2*N , 2*N , 0, 360, 0);
		gp.fill(ball); //Makes the visualise-objects visible
	    }
	    
	    //Draw own team
	    gp.setPaint(Color.red); //in red
	    PlayerListElement ple = null;
	    Iterator it = tm.iterator(); //create iterator
	    while ( it.hasNext() ) //is there a next element?
	    {
		ple = (PlayerListElement)it.next(); //get next element
		//get AbsPosition from relative Position
		CVector tmi = ple.position.toCVector(player.getDirection()).add(p); 
		Arc2D teammember = new Arc2D.Double(dx+N*(tmi.x-0.5), dy+N*(tmi.y-0.5), N, N, 0, 360, 0);
		gp.fill(teammember); //Makes the visualise-objects visible
	    }

	    //Draw the enemy
	    gp.setPaint(Color.blue); //in blue
	    it = op.iterator();//create iterator
	    while ( it.hasNext() )  //is there a next element?
	    {
		ple = (PlayerListElement)it.next(); //get next element
		//get AbsPosition from relative Position
		CVector opi = ple.position.toCVector(player.getDirection()).add(p);
		Arc2D opponents = new Arc2D.Double(dx+N*(opi.x-0.5), dy+(N*opi.y-0.5), N, N, 0, 360, 0);
		gp.fill(opponents); //Makes the visualise-objects visible
	    }
	}
}
