
import java.lang.Math;

/**
 * as this is a kind of structure, the attributes are public<BR>
 * the direction angle is stored in rad
 *
 * @author Matthias Lang
 * @version $Id: PVector.java,v 1.1 2002/07/10 00:37:01 Lang Exp $
 **/
public class PVector
{
    double distance;
    int direction;
    
    /** 
     * creates a new polar vector initializing distance and direction to zero
     **/
    public PVector()
    {
	distance = 0.;
	direction = 0;
    }

    /**
     * Creates a new polar vector, initializing distance and direction to the given values.
     * transforms diraction angle into rad
     *
     * @param distance initial value for distance
     * @param direction initial value for direction in degrees
     **/
    public PVector( double distance, int direction )
    {
	this.direction = direction;
	this.distance = distance;
    }

    /**
     * transforms the polar vector to a cartesian vector
     *
     * @param lookAtAngle absolute look-at angle (BodyAngle + HeadAngle)
     * @return the new cartesian vector
     **/
    public CVector toCVector(int lookAtAngle)
    {
	return  new CVector(distance*Math.cos(Math.PI/180.*(direction+lookAtAngle)),
			    distance*Math.sin(Math.PI/180.*(direction+lookAtAngle)) );
    }

    /**
     * calculates the distance between two seen objects
     *
     * @param other another polar vector (the other seen object)
     * @return the distance
     **/
    public double diff( PVector other )
    {
	return Math.sqrt(distance*distance + other.distance*other.distance
			 - 2*distance*other.distance*
			 Math.cos(Math.PI/180*(direction - other.direction)) );
    }
}
