/**
 * This is the central class of one player.<BR>
 * It sends the initial message and starts the receiver and strategy thread.
 *
 * @author Matthias Lang
 * @version 1.2
 **/
public class Player extends Thread
{

    private CommunicationChannel communication;
    private Worldmodell data;
    private DivideString messageHandler;
    private Strategy strategy;

    /**
     * Creates a new <code>Player</code> instance.
     *
     * @param servername a <code>String</code> value specifying the the name of the server
     * @param serverport an <code>int</code> value specifying the port on which the server
     *                   listens for incoming packets
     * @param teamName a <code>String</code> value is the name of the team you want to join/create
     **/
    public Player( String servername, int serverport, String teamName, int playerNum )
    {
        communication = new CommunicationChannel(servername, serverport);

	data = new Worldmodell();

	if (playerNum == 1)
	{
	    communication.send("(init "+teamName+" (version 7.0) (goalie)");
	    strategy = new StrategyKeeper(communication,data);
	}
	else
	{
	    communication.send("(init "+teamName+" (version 7.0))");
	    if (playerNum < 9) strategy = new StrategyDefender(communication,data);
	    else strategy = new StrategyForward(communication,data);
	}
	//strategy.start();
	messageHandler = new DivideString( teamName, data, strategy );
    }


    /**
     * This method had to be overwritten from <code>Thread</code>. It is
     * executed when <code>start()</code> is called for the <code>Player</code>
     * Thread
     **/
    public void run()
    {
	while (!isInterrupted())
	    mainLoop();

	System.out.println("Player "+data.getPlayerNum()+" is going down...");
	System.out.println("\t\t\tdone.");
    }

    /**
     * main loop which is executed from run
     **/
    private void mainLoop()
    {
	messageHandler.runDivideString( communication.receive() );
    }

    /**
     * returns pointer to Worldmodell class
     *
     * @return object of type Worldmodell containing all data
     **/
    public Worldmodell getData()
    {
	return data;
    }

}













