import java.util.*;

/**
 * StrategyDefender class which calculates positions and decides what to do next
 * methods in which forward and defender take different decisions
 * @author Andrea Barth
 * @author Matthias Lang
 * @version $Id: StrategyDefender.java,v 1.4 2002/07/15 01:18:43 Lang Exp $
 **/




public class StrategyDefender extends Strategy
{
   
    private double startx, starty, endx, endy;
    private int side;
    /**
     * the constructor initalizes a Strategy Defender object equals Strategy object
     *
     **/
    public StrategyDefender(CommunicationChannel com, Worldmodell data)
    {
	super(com, data);
    }


    /**
     * calculates the kick, where to kick and executes it
     *
     **/
    protected void calcKick()
    {
	int sideToShoot;
	if (side == 0) {
	    if (ballSpeed.y>0) sideToShoot = 1;
	    else sideToShoot = -1;
	}
	else sideToShoot = side;

	LinkedList teamMembers = data.getTeamMembers();
	Iterator it = teamMembers.iterator();
	
	while (it.hasNext())
	{
	    PlayerListElement potential = (PlayerListElement)it.next();
	    CVector tempPos = potential.position.toCVector(absAngle).add(ownPosition);
	    //search for team member on my side who is significantly nearer to goal
	    if ( side*tempPos.y > 20 && tempPos.x > ownPosition.x+10 ) {
		action.kick( 100., potential.position.direction ); 
		return;
	    }
	}
	//play over wings
	if (ownPosition.x<-15) action.kick( 100., calcAngleToPos( new CVector( 0., side*15 )) );
	// shoot on goal if no other target found
	else action.kick( 100., calcAngleToPos(GOALPOSITION) );
    
    }

    /**
     * decides if the player runs towards the ball
     *
     * @return boolean: true if the player should run to the ball, else false
     **/
    protected boolean decisionRunToBall()
    {
	return ( checkInside(absBallPosition) || !data.otherPlayerNext2Ball() );
    }

    /**
     * adjusts player's position inside his assigned area
     **/
    protected void adjustPosition()
    {
	CVector target;
	if (absBallPosition.x>10 && playerNum>5)
	    target = standPosition.add( new CVector(10., 0.));
	else target = standPosition;

	if (Math.abs(ownPosition.sub(target).length())<3)
	    turnToBall();
	else 
	    calcDashPowerAndTurn(standPosition);
	

    }

    
    /**
     * gives his standard and start position to each player
     *
     * @author Annika Niehage
     **/
    protected void  calcStandPosition()
    {
	if(playerNum == 2)
	{    
	    standPosition = new CVector(-32.5, -19.5);
	    startPosition = new CVector(-32.5, -19.5);
	    setArea( -52.5, -34, -15, -15 );
	    side = -1;
	}

	if(playerNum == 3 || playerNum == 1)
	{    
	    standPosition = new CVector(-25, -5);
	    startPosition = new CVector(-25, -5);
	    setArea( -52.5, -18, -15, 4 );
	    side = 0;
	}

	if(playerNum == 4)
	{    
	    standPosition = new CVector(-32.5, 5);
	    startPosition = new CVector(-32.5, 5);
	    setArea( -52.5, -4, -15, 18 );
	    side = 0;
	}

	if(playerNum == 5)
	{
	    standPosition = new CVector(-32.5, 19.5);
	    startPosition = new CVector(-32.5, 19.5); 
	    setArea( -52.5, 15, -15, 34 );
	    side = 1;
	}
	if(playerNum == 6)
	{    
	    standPosition = new CVector(-5, -15);
	    startPosition = new CVector(-20, -15);
	    setArea( -20, -34, 15, -19 );
	    side = -1;
	}

	if(playerNum == 7)
	{    
	    standPosition = new CVector(-12, 0);
	    startPosition = new CVector(-2,  0);
	    setArea( -20, -12, 15, 15 );
	    side = 0;
	}
	if(playerNum == 8)
	{    
	    standPosition = new CVector(-5, 15);
	    startPosition = new CVector(-20, 15);
	    setArea( -20, 19, 15, 34 );
	    side = 1;    
	}

    }

    /**
     * defines an area in which player should act
     */
    private void setArea( double fstartx, double fstarty, double fendx, double fendy )
    {
	startx = fstartx;
	starty = fstarty;
	endx = fendx;
	endy = fendy;
    }

    /**
     * checks if given position is inside player's area
     * @param CVector position to be checked
     * @return true if inside area, false otherwise
     **/
    private boolean checkInside( CVector pos )
    {
	return ( pos.x >= startx && pos.x <= endx && pos.y >= starty && pos.y <= endy );
    }

}

