import java.math.*;
import java.util.*;

/**
 * StrategyKeeper class which implements the Keepers actions  and decides what to do next
 *
 * @author Alexander Mueller
 * @author Oliver Wagner
 * @version $Id: StrategyKeeper.java,v 1.3 2002/07/14 22:22:12 Wagner Exp $
 **/

public class StrategyKeeper extends Strategy
{
   
    private final double CATCHABLEDISTANCE = 1.9;
    private final CVector GOALEDGE = new CVector(-52.5 , 7.01 );
    private final int KICKAWAYANGLE = 25;
    private final double KEEPERLINE = -51.5;
    private final CVector STANDPOSITION = new CVector (KEEPERLINE, 0.);
    private final int  ANGLETOLERANCE = 15;
    private CVector usefulBallMeetingPoint;
    private final CVector RIGHTFORWARD = new CVector (15, 20) , LEFTFORWARD = new CVector(15, -20);
    private boolean goalKick;
    private boolean goalieCatchBall;
    private boolean catchNextPossibility = false;
    private int lastCatchCycle = -5;
    private CVector ourOldBallSpeedVector = new CVector (0. ,0.);
    private boolean oneTimeBeforeDidSeeBall = false;


    /** 
    * This is the constructor.
    * 
    * It calls the super-class-constructor.
    * @param com The communication channel for the super class.
    * @param data The reference on the player data. 
    * @author Oliver Wagner, Alexander Mueller
    **/

    public StrategyKeeper(CommunicationChannel com,Worldmodell data)
    {
	super(com, data);
	goalKick = false;
	goalieCatchBall = false;
    }


    /** 
    * This methode checks the playstate and decides what action-methode should be called
    * 
    * @author Oliver Wagner, Alexander Mueller
    **/

    public void decide() {
	// Es kam oefters vor, dass wir nicht definierte Werte bekamen. Diese entstanden bei einem Teilen durch
	// 0. Diese koennen wir nicht verwenden und setzen sie auf moeglichst gute Werte.
	if(ballSpeed != null && Double.isNaN(ballSpeed.x)) {
	    ballSpeed.x = 0.0000000000001;
	    if (ballMeetingPoint != null && Double.isNaN(ballMeetingPoint.x))
		ballMeetingPoint.x = absBallPosition.x;
	}
	
	// Wenn wir den ersten Zyklus den Ball nicht sehen (also im Zyklus davor haben wir ihn gesehen), dann 
	// verwenden wir die alte Ballposition als die
	// aktuelle. Somit haben wir einen Buffer, wenn einmal nicht der Ball erkannt wird. Ab dem zweiten 
	// Zyklus ball nicht sehen ist "oneTimeBeforeDidSeeBall" false => die aktuelle Ballposition wird 
	// nicht uebernommen. Sobald einmal ballPosition != null ist, also der Ball gesehen wurde, wird 
	// oneTimeBeforeDidSeeBall auf true gesetzt, da wir ihn jetzt wieder sehen.
	if (ballPosition == null) {
	    if (oneTimeBeforeDidSeeBall) {
		ballPosition = oldBallPosition;
		absBallPosition = ownPosition.add(ballPosition.toCVector(absAngle));
		ballSpeed = new CVector (0., 0.);
		// Berechnet die ballFuturePosition mit dem alten Wert
		if (ballSpeed.length() < 0.5)
		    ballFuturePosition = absBallPosition;
		else
		    ballFuturePosition = new CVector (absBallPosition.x + ballSpeed.x * DECAYBALL,
 						      absBallPosition.y + ballSpeed.y * DECAYBALL);
		calcBallMeetingPoint();
	    }
	}	    
	else
	    oneTimeBeforeDidSeeBall = true;
	
	
	// hier werden die angles auf nuetzliche Werte korrigiert.
	calcUsefullAngles();
	calcUsefulBallMeetingPoint();

	// entsprechend des PlayModes wird gehandelt.
	switch (data.getPlayMode()) {
	    case 0:
		playOn();
		break;
	    case 1:
		executeGoalKick();
		break;
	    case 2:
	    case 3:
	    case 4:
		playOn();
		break;
	    case 5:
		executeFreeKickInPenaltyArea();
		break;
	    case 6:
	    case 7:
	    case 8:
	    case 9:
	    case 10:
	    case 11:
	    case 12:
		playOn();
		break;
	    case 13:
		handleLogOut();
		break;
	    case 14:
	    case 15:
	    case 16:
	    case 17:
		playOn();
		break;
	    default:
		playOn();
		break;
	}

	if (ballPosition == oldBallPosition)
	    oneTimeBeforeDidSeeBall = false;
    }


    /** 
     * This methode is the methode for the running game.
     * 
     * In it the decisions for the running game are made.
     * @author Oliver Wagner, Alexander Mueller
     **/ 

    private void playOn () {
	if (ballPosition == null) {
	    searchBall();
	}
	else if ((ballPosition.distance <= CATCHABLEDISTANCE || catchNextPossibility)
		 && (timeCycle - lastCatchCycle > 5)) {
	    if (currentData) {
		action.caught(ballPosition.direction-neckAngle);
	    }
	    else {
		action.caught(-neckAngle);
	    }
	    lastCatchCycle = timeCycle;
	    catchNextPossibility = false;
	} 
	else if (ballFuturePosition.sub(ownPosition).length() <= CATCHABLEDISTANCE 
		 && (timeCycle - lastCatchCycle > 4)) {
	    turn(calcAngleToPos(ballFuturePosition) - neckAngle);
	    turnNeck(-neckAngle);
	    catchNextPossibility = true;
	}
	else if (ballPosition.distance <= KICKABLEDISTANCE)
	    kickToPosition(absBallPosition);
 	else if (inCruisingRadius(absBallPosition))
 	    runToPositionAndKeepSeeingBall(usefulBallMeetingPoint);
 	else if (inCruisingRadius(usefulBallMeetingPoint))
 	    runToPositionAndKeepSeeingBall(usefulBallMeetingPoint);
	else if (STANDPOSITION.sub(ownPosition).length() > 2 ) {
	    runToPositionAndKeepSeeingBall(STANDPOSITION);
	}	
	else { 
            turnFeetAndNeckToBall();
	}
    }


    /** 
     * This methode calculates a meetingBall that is in the field.
     *  
     * When a point is behind the field, the latest position in the field is given back.
     * The methode is not very exact. This is a follow from the unexact given data.
     * @author Oliver Wagner, Alexander Mueller
     **/

    private void calcUsefulBallMeetingPoint () {
	 if (ballMeetingPoint.x >= KEEPERLINE)
	     usefulBallMeetingPoint = ballMeetingPoint;
	 else
	     usefulBallMeetingPoint = calcPointOfIntersection();
     }

    /**
     * I forgot what this methode does!
     *
     * One Moment, I remember! I can not belive, but this method "calcPointOfIntersection" 
     * calculates the point of intersection between KEEPERLINE and ballSpeed-Vector.
     *
     * @author Oliver Wagner, Alexander Mueller
     * @return Oh god, I forgot again ...
     **/

    private CVector calcPointOfIntersection () {
	if (ballSpeed.x == 0)
	    return new CVector (KEEPERLINE, absBallPosition.y);
	else {
	    CVector pointOfIntersection = new CVector (KEEPERLINE , absBallPosition.y + ballSpeed.y * 
						   Math.abs((KEEPERLINE - absBallPosition.x) / ballSpeed.x));
	    if (pointOfIntersection.y > GOALEDGE.y)
		pointOfIntersection.y = GOALEDGE.y;
	    else if (pointOfIntersection.y < - GOALEDGE.y)
		pointOfIntersection.y = - GOALEDGE.y;
	    return pointOfIntersection;
	}
    }


    /** 
     * This methode is the keepers main running methode.
     *
     * The keeper turns to the given position and keeps his eyes on the ball.
     * When the angle between his aim and the ball gets above 110 degrees he turns
     * and runs backwards. 
     * @param position The position where the keeper shall run to.
     * @author Oliver Wagner, Alexander Mueller
     **/

    private void runToPositionAndKeepSeeingBall(CVector position) {
	//Winkel zwischen Blickwinkel und Ziel(=position)
	int angleToPos = calcAngleToPos(position);

	// Wenn aim schon sehr nah ist! Nur beim vorwaertslaufen!!!
	if (Math.abs(angleToPos + neckAngle) > 5 && ballPosition.distance < 5 && 
	    Math.abs(angleToPos + neckAngle) < ANGLETOLERANCE) {
	    turn(angleToPos + neckAngle);
	    turnNeck(ballPosition.direction - angleToPos - neckAngle);
	}
	else if (ballPosition.distance < 1.5 && Math.abs(angleToPos + neckAngle) < 5) {
	    turnNeck(ballPosition.direction);
	    action.dash(20);
	}


	// wenn aim in der bodyAngle-toleranz ist und wir somit beschleunigen wollen.
	else if (Math.abs(angleToPos + neckAngle) < ANGLETOLERANCE) {
	    turnNeck(ballPosition.direction);
	    action.dash(100);
	}
 	else if (Math.abs(angleToPos + neckAngle) > (180-(ANGLETOLERANCE))) {
 	    turnNeck(ballPosition.direction);
 	    action.dash(-100);
 	}

	// richtig die Fuesse auf die aim Position ausrichten und dabei den Ball im Blick behalten.

	// ausrichten um vorwaerts zum Ball zu laufen
	else if (Math.abs( - ballPosition.direction + angleToPos) <= 110) {
	    turn(angleToPos + neckAngle);
	    turnNeck(ballPosition.direction - (angleToPos + neckAngle));
	}
	// ausrichten um rueckwaerts zum Ball zu laufen
	else if (Math.abs(- ballPosition.direction + angleToPos) > 90) {
	    int meinTurn;
	    if (angleToPos < 0)
		meinTurn = 180 - Math.abs(angleToPos - neckAngle);
	    else
		meinTurn = - (180 - Math.abs(angleToPos - neckAngle));
	    turn (meinTurn);
	    turnNeck (ballPosition.direction - meinTurn - neckAngle);
	}
    }


    /** 
     * This methode is used to check if a given position is within penalty area.
     * 
     * @param position The position that has to be tested
     * @author Oliver Wagner, Alexander Mueller
     * @return if in Cruising Radius -> true; else -> false
     **/

    private boolean inCruisingRadius(CVector position) {
        if (position.x < -36.5 && position.x > KEEPERLINE && (Math.abs(position.y) < 19.5))
	    return true;
	else 
	    return false;
    }


    /** 
     * This methode is used to ensure that all angles are smaller than 180 degrees
     * 
     * @author Oliver Wagner, Alexander Mueller
     **/

    private void calcUsefullAngles () {

	bodyAngle = bodyAngle % 360;
	if (bodyAngle > 180)
	    bodyAngle = -360 + bodyAngle;
	else if (bodyAngle < -180)
	    bodyAngle = 360 + bodyAngle;

	if (neckAngle < -90)
	    neckAngle = -90;
	else if (neckAngle > 90)
	    neckAngle = 90;

	absAngle = absAngle % 360;
	if (absAngle > 180)
	    absAngle = -360 + absAngle;
	else if (absAngle < -180)
	    absAngle = 360 + absAngle;
    }


    /** 
     * This methode is called to turn the player complete to the ball.
     * 
     * @author Oliver Wagner, Alexander Mueller
     **/

    private void turnFeetAndNeckToBall() {
        turn(neckAngle + ballPosition.direction);
        // = absAngle + ballPosition.direction - bodyAngle
        turnNeck(-neckAngle);
    }


    /** 
     * This methode is used to turn the goalies neck. 
     * 
     * The angles get corrected.
     * @param degrees The angle to turn the neck.
     * @author Oliver Wagner, Alexander Mueller
     **/

    private void turnNeck(int degrees) {
	action.turn_neck(degrees);
	neckAngle += degrees;
	absAngle += degrees;
    }


    /** 
     * This methode is used to turn.
     * 
     * The methode from Strategy class cannot be used, because the goalie turns his neck so that
     * the angles have to be corrected.
     * @param degrees The angle that has to be turned.
     * @author Oliver Wagner, Alexander Mueller
     **/

    private void turn(int degrees) {
	action.turn(degrees);
	bodyAngle += degrees;
	absAngle += degrees;
    }


    /** 
     * This methode is used to search the ball.
     *
     * The methode of Strategy class cannot be used, because the keepers - different from the other
     * players - turns his neck.
     * @author Oliver Wagner, Alexander Mueller
     **/

    private void searchBall () {
	if (neckAngle != 0) {
	    turnNeck(-neckAngle);
	}

	if (currentData)
	    if (oldBallPosition.direction > 0)
		turn(20);
	    else
		turn(-20);
    }


    /** 
     * This methode is used to kick the ball to a specified position.
     * 
     * The kick is normally not to exact.
     * 
     * @param aim The position where the ball should be kicked to.
     * @author Oliver Wagner, Alexander Mueller
     **/

    private void kickToPosition(CVector aim) {
	//Winkel zwischen Blickwinkel und Ziel(=position)
	int angleToPos = calcAngleToPos(aim);
	action.kick(100, (angleToPos + neckAngle));
    }


    /** 
     * This methode is used to execute a goal kick.
     * 
     * @author Oliver Wagner, Alexander Mueller
     **/
    
    private void executeGoalKick() {
	if (ballPosition == null)
	    searchBall();
	else if (ballPosition.distance >= KICKABLEDISTANCE)
	    runToPositionAndKeepSeeingBall(absBallPosition);
	else if (absBallPosition.y > 0)
	    kickToPosition(RIGHTFORWARD);
	else if (absBallPosition.y <= 0)
	    kickToPosition(LEFTFORWARD);
	else runToPositionAndKeepSeeingBall(STANDPOSITION);
    }


    /** 
     * This methode implements the execution of freekicks in the penalty area,especially 
     * when the goalie cougth the ball.
     * 
     * @author Oliver Wagner, Alexander Mueller
     **/
    
    private void executeFreeKickInPenaltyArea() {
	if (ballPosition == null)
	    searchBall();
	else if (inCruisingRadius(absBallPosition))
	    executeGoalKick();
    }


    /** 
     * This methode is thougth to handle the logout.
     * 
     * @author Oliver Wagner, Alexander Mueller
     **/ 

    private void handleLogOut() {}



    
    // initialize the unneeded abstract methods
    protected boolean decisionRunToBall() {return true;}
    protected void adjustPosition(){}
    protected void calcStandPosition() {
	startPosition = STANDPOSITION;
	standPosition = STANDPOSITION;
    }
    protected void calcKick() {}



    
    /** 
     * This Methode calcs the ballMeetingPoint like in Calculation class,
     * but the meetingPoint is calculated 60 cycle into the future.
     * 
     * @author Oliver Wagner, Alexander Mueller
     **/

    protected void calcBallMeetingPoint()
    {
 	if (ballPosition.distance > 10)	{
	    ballMeetingPoint.x = absBallPosition.x + ballSpeed.x * 16.63241875;
	    // (1- Math.pow(DECAYBALL, 100))/(1-DECAYBALL)
	    ballMeetingPoint.y = absBallPosition.y + ballSpeed.y * 16.63241875;
            // (1- Math.pow(DECAYBALL, 100))/(1-DECAYBALL)
 	}
 	else {
 	    if (ballPosition.distance > 5)
		ballMeetingPoint = ballFuturePosition;
	    else
		ballMeetingPoint = absBallPosition;
	}
    }
}
