import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.lang.Runnable;

/**
 *This class is the main Visualisation-class.
 *
 *By calling the Constructor with the correct Worldmodell Data Array
 *a visualisation is started.
 *
 *@author Alexander Mller
 *@author Olga Mirkina
**/

public class WBV extends JFrame implements Runnable  
{
    private static int selectedR = 11;    //Saves the Player number, default: No player
    public double enlarge = 8;            //Enlargment factor of the field
    public Worldmodell[] PlayerList = new Worldmodell[11]; //Creates an array with players' information
    public Draw Spielfeld = new Draw(enlarge);   //Creates the Field-Object where the visualistion takes place
    JPanel Auswahl = new JPanel();   //creates an empty JPanel for the Buttons
     
    /**
     *This is the Constructor of WBV.java
     *
     *Because of WBV extending JFrame this constructor creates a new JFrame, set the PlayerList to the given Data
     *and calls fuelle() for further control of the JFrame
     *
     *@param Worldmodell[] team which is an 11-parts-Array that contains players data.
     **/ 
    public WBV(Worldmodell[] team)
	{
	    fuelle();
	    PlayerList = team;
	}
    
    /** 
     *This method draws the outer Main Window. 
     *
     *This method gives the Main Window its components: the Selection Buttons in a JPanel that's filling methode 
     *is called as well and the Playground in a Draw-Object.
     **/    
    public void fuelle()
	{
	    fillHauptfenster();
	    fillAuswahl(Auswahl);
	    setSize(880,544);
	    setTitle("WBV");
	    getContentPane().add(Auswahl, BorderLayout.WEST);
	    getContentPane().add(Spielfeld, BorderLayout.CENTER);
	    setVisible(true);
     }

    /**
     *This methode compares the size of the window and its minimum size to ensure
     *that the window always stays bigger than the minimum size.
     *
     *@param Dimension is Here the methode gets the actual size.
     *@param Dimension min Here the methode gets its minimum size.
     *@return Dimension is The methode returns a possible size. (Smaller than min isn't allowed.)
     **/

private Dimension Compare(Dimension is,Dimension min)
{
	if (is.width<(min.width)) is.width = min.width;
	if (is.height<(min.height+10)) is.height = min.height+10;
	return is;
}


    /**
     *This method supplements the Main Window. 
     *
     *This method gives the Main Window a BorderLayout, a Closing-the-Window operation and it's behaviour if
     *it's resized.
     **/

    public void fillHauptfenster()
	{
	    
	    //Quits the visualization if the window is closed
	    addWindowListener(new WindowAdapter()
		{
		    public void windowClosing(WindowEvent e)
			{
			    dispose();
			}
		});
	    getContentPane().setLayout(new BorderLayout());
	    //tests the size of the main window and resizes it
	    addComponentListener(new ComponentAdapter()
		{
		    public void componentResized(ComponentEvent ce)
			{
			    Dimension s = getSize();
			    Dimension d = getMinimumSize();
			    s = Compare (s,d);
			    setSize(s);
			    double horChng = 850/(double)(s.width-200);
			    double verChng = 680/(double)(s.height);
			    if(horChng>verChng) enlarge = 8/horChng;
			    else enlarge = 8/verChng;
			    Spielfeld.setFactor(enlarge);
			}
		});
	 
     }

    /**
     *This method creates and manages the Selection List.
     *
     *This method creates a Selection List as JRadioButtons, ordered in a
     *ButtonGroup to guarantee that only one state is possible at a time. Each
     *button has an ItemListener which activates if the button is pushed and sets the
     *Chosen-Player-Data in the Draw-Object
     *
     *@param Auswahl The JPanel to be filled
     **/

    public void fillAuswahl(JPanel Auswahl)
	{
	    Auswahl.setSize(544,200);
	    Auswahl.setLayout(new GridLayout(13,1)); //Makes all buttons be in one column
	    Auswahl.add(new JLabel("Choice"));  //adds a headline
	    ButtonGroup bg = new ButtonGroup();
	    final JRadioButton[] r =
	    {
		new JRadioButton(" Player 1"),
		new JRadioButton(" Player 2"),
		new JRadioButton(" Player 3"),
		new JRadioButton(" Player 4"),
		new JRadioButton(" Player 5"),
		new JRadioButton(" Player 6"),
		new JRadioButton(" Player 7"),
		new JRadioButton(" Player 8"),
		new JRadioButton(" Player 9"),
		new JRadioButton(" Player 10"),
		new JRadioButton(" Player 11"),
		new JRadioButton(" No Player")
		    };
	    //Here the ItemListener is created
	    //It's declarated by an inner class that tells the listener
	    //how to react on an StateChangeEvent which is released when
	    //a button is chosen.
	    ItemListener listener= new ItemListener()
		{
		    public void itemStateChanged(ItemEvent e)
			{
			    if(e.getStateChange()==ItemEvent.SELECTED)
			    {
				int i = 0;
				//this while-loop test which button is selected
				while(!r[i].isSelected())
				{
				    i++;
				}
				setSelectedR(i);   // sets the selectedR Variable
				if (i>=0) 
				    if(i<11){ 
					//sets the Playerdata and the Playernumber in the Draw-Object
					Spielfeld.setPlayer(PlayerList[i],i); 
					}
				    else Spielfeld.setPlayer(11);	//sets the Draw-Object to "No Visualisation"
				else
				    Spielfeld.setPlayer(11);   //sets the Draw-Object to "No Visualisation"
			    }
			}
		};
	    for(int i = 0; i < r.length; i++)   //for every button do
	    {
		r[i].addItemListener(listener); //give the button an ItemListener
		bg.add(r[i]);  //add the button to the buttongroup
		Auswahl.add(r[i]);  //add the button to The JPanel Auswahl
	    }
	}

    /**
     *This method is the "Setter" for the selectedR variable.
     *
     *@param i Through this variable will be passed the value for selectedR
     **/

    public void setSelectedR(int i)
	{
	    selectedR = i;
	}

    /**
     *This method is the "Getter" for the selectedR variable.
     *
     *@return selectedR
     **/

    public int getSelectedR()
	{
	    return selectedR;
	}
    
    /**
     *This method serves the permanent visualization of the selected player.
     *
     *This method calls the repaint-methode that calls paint() for the whole JPanel.
     *This methode is the run methode of the WBV-Thread in Robocupteam.java
     **/

    public void run()
	{
	    while(true)
	    {
		repaint();
		try
		{
		    Thread.sleep(100);
		}
		catch(InterruptedException e){}
	    }
	}
}
