import java.util.*;

/**
 * This is the data class where the main data is stored:
 * ownPosition, ballPosition, information from see and hear messages
 *
 * @author Matthias Lang
 * @author Annika Niehage
 * @author Tim Oliver Wagner
 * @version $Id: Worldmodell.java,v 1.5 2002/07/14 19:37:34 Niehage Exp $
 **/
public class Worldmodell
{
    private CVector ownPosition, ownSpeed, ballSpeed = new CVector();
    private PVector ballPosition, ballChange;
    private Flag flag;

    /* Linked lists with elements of type PlayerListElement */
    private LinkedList teamMembers, opponents, tempTeamMembers, tempOpponents;
    private int time, playDir, playerNumber;
    private int neckAngle, bodyAngle, direction;

    private int otherStart = 0, otherEnd = 0;
    private boolean otherEmpty = true;
    private int playMode = 14;

    public boolean newSeeMsg = false, seeSomething = false;

    /**
     * Constructor
     **/    
    public Worldmodell()
    {
	ownPosition = new CVector();
	ownSpeed = new CVector();
	tempTeamMembers = new LinkedList();
	tempOpponents = new LinkedList();
	playerNumber = 0; // value before first see message; is tested by Strategy
	direction = 0;
	neckAngle = 0;
	bodyAngle = 0;
    }

    /**
     * Set the last HearMessage
     **/
    public void setHearMessage(String[] aMessage) {
	if (aMessage[1].equalsIgnoreCase("referee"))
	    setPlayMode(aMessage);
    }

    /**
     * set the correct current play Mode
     *
     * @author Alexander Mueller
     * @author Oliver Wagner
     *
     * @param aMsg the refere mesage
     **/
    private void setPlayMode (String[] string) 
	{
	if (string[2].startsWith("play_on"))
	    playMode = 0;
	else if ((string[2].startsWith("goal_kick_l") && getPlayDir() == 1) || 
		 (string[2].startsWith("goal_kick_r") && getPlayDir() == -1))
	    playMode = 1;
	else if ((string[2].startsWith("goal_kick_l") && getPlayDir() == -1) || 
		 (string[2].startsWith("goal_kick_r") && getPlayDir() == 1))
	    playMode = 2;
	else if ((string[2].startsWith("kick_in_l") && getPlayDir() == 1) || 
		 (string[2].startsWith("kick_in_r") && getPlayDir() == -1))
	    playMode = 3;
	else if ((string[2].startsWith("kick_in_l") && getPlayDir() == -1) || 
		 (string[2].startsWith("kick_in_r") && getPlayDir() == 1))
	    playMode = 4;
	else if ((string[2].startsWith("free_kick_l") && getPlayDir() == 1) || 
		 (string[2].startsWith("free_kick_r") && getPlayDir() == -1))
	    playMode = 5;
	else if ((string[2].startsWith("free_kick_l") && getPlayDir() == -1) || 
		 (string[2].startsWith("free_kick_r") && getPlayDir() == 1))
	    playMode = 6;
	else if ((string[2].startsWith("kick_off_l") && getPlayDir() == 1) || 
		 (string[2].startsWith("kick_off_r") && getPlayDir() == -1))
	    playMode = 7;
	else if ((string[2].startsWith("kick_off_l") && getPlayDir() == -1) || 
		 (string[2].startsWith("kick_off_r") && getPlayDir() == 1))
	    playMode = 8;
	else if ((string[2].startsWith("goal_l") && getPlayDir() == 1) || 
		 (string[2].startsWith("goal_r") && getPlayDir() == -1))
	    playMode = 9;
	else if ((string[2].startsWith("goal_l") && getPlayDir() == -1) || 
		 (string[2].startsWith("goal_r") && getPlayDir() == 1))
	    playMode = 10;
	else if ((string[2].startsWith("corner_kick_l") && getPlayDir() == 1) || 
		 (string[2].startsWith("corner_kick_r") && getPlayDir() == -1))
	    playMode = 11;
	else if ((string[2].startsWith("corner_kick_l") && getPlayDir() == -1) || 
		 (string[2].startsWith("corner_kick_r") && getPlayDir() == 1))
	    playMode = 12;
	else if (string[2].startsWith("timer_over"))
	    playMode = 13;
	else if (string[2].startsWith("before_kick_off") ||
		 string[2].startsWith("half_time") )
	    playMode = 14;
	else if (string[2].startsWith("drop_ball"))
	    playMode = 15;
	else if ((string[2].startsWith("offside_l") && getPlayDir() == 1) || 
		 (string[2].startsWith("offside_r") && getPlayDir() == -1))
	    playMode = 16;
	else if ((string[2].startsWith("offside_l") && getPlayDir() == -1) || 
		 (string[2].startsWith("offside_r") && getPlayDir() == 1))
	    playMode = 17;
    }	


    /**
     * This methode returns an integer as a sign for the game-mode
     *
     * 0 : play_on
     * 1 : goal_kick_forUs
     * 2 : goal_kick_enemy
     * 3 : kick_in_forUs
     * 4 : kick_in_enemy
     * 5 : free_kick_forUs
     * 6 : free_kick_enemy
     * 7 : kick_off_forUs
     * 8 : kick_off_enemy
     * 9 : goal_forUs
     * 10: goal_enemy
     * 11: corner_kick_forUs
     * 12: corner_kick_enemy
     * 13: time_over
     * 14: before_kick_off
     * 15: drop_ball
     * 16: offside_forUs
     * 17: offside_enemy
     *
     * @author Alexander Mueller
     * @author Oliver Wagner
     *
     * @return the current play mode
     **/
    public int getPlayMode () {
	return playMode;
    }

    /**
     * sets the time cycle
     *
     * @param time new timecycle
     **/
    public void setTimeAndPlayerLists( int currentTime )
    {
	this.time = currentTime;
	// copy lists
	teamMembers = tempTeamMembers;
	opponents = tempOpponents;
	tempTeamMembers = new LinkedList();
	tempOpponents = new LinkedList();
    }

    /**
     * returns the timecycle
     *
     * @return the timecycle
     **/
    public int getTime()
    {
	return time;
    }

    /**
     * returns 1 for left to right and -1 otherwise
     *
     * @return see above
     **/
    public int getPlayDir()
    {
	return playDir;
    }

    /**
     * sets initial values, received after init message
     * direction in which we play is +1 for left to right and -1 otherwise
     *
     * @param dirEtNum String containing l or r and the player number (1-11)
     **/
    public void setInitValues( String dirEtNum )
    {
	if (dirEtNum.charAt(6)=='l') playDir = 1;
	else playDir = -1;
	if (dirEtNum.charAt(9)==' ') playerNumber = dirEtNum.charAt(8)-48;
	else playerNumber = dirEtNum.charAt(9)-38; // 38 because it is >= 10
    }

    /**
     * returns the player's number
     *
     *@return what do you think is that? can you guess? maybe the player's weight?
     **/
    public int getPlayerNum()
    {
	return playerNumber;
    }

    /**
     * sets the global lookAtAngle on the basis of a seen line
     *
     * @param data String array containing which line and the angle to it
     **/
    public void setDirection( String[] data )
    {
	char lineType = data[0].charAt(2);
	int angle = Integer.parseInt( data[2] );

	int sign = 0;
	if (lineType == 't' || lineType == 'b')
	{
	    if (lineType=='t') sign = -playDir;
	    else sign = playDir;
	    if (sign < 0)
	    {
		if (angle < 0) direction = -180 - angle;
		else direction = -angle;
	    }
	    else
	    {
		if (angle < 0) direction = -angle;
		else direction = 180 - angle;
	    }
	}
	else
	{
	    if (lineType=='l') sign = -playDir;
	    else sign = playDir;
	    if (sign < 0)
	    {
		if (angle < 0) direction = 90 - angle;
		else direction = - 90 - angle;
	    }
	    else
	    {
		if (angle < 0) direction = - 90 - angle;
		else direction = 90 - angle;
	    }
	}
	bodyAngle = direction - neckAngle;
    }

    /**
     * returns direction ( bodyAngle + neckAngle )
     *
     * @return direction
     **/
    public int getDirection()
    {
	return direction;
    }

    /**
     * sets the neck angle, also called head angle,
     * and calculates the body angle
     * <I>neckAngle is still unused</I>
     *
     * @param angle a String containing the value
     **/
    public void setNeckAngle( int angle )
    {
	neckAngle = angle;
	bodyAngle = direction - neckAngle;
    }

    /**
     * returns the neck angle
     *
     * @return you won't believe it, but it's the neck angle
     **/
    public int getNeckAngle()
    {
	return neckAngle;
    }

    /**
     * returns the body angle
     *
     * @return hm, i've forgotten what it is
     **/
    public int getBodyAngle()
    {
	return bodyAngle;
    }

    /**
     * calculates own position and speed from the relative position of a flag
     * and the absolute direction
     **/
    public void calcOwnPosition()
    {
	CVector temp = ownPosition;
	ownPosition = flag.absPos.sub( flag.relPos.toCVector(direction) );
	ownSpeed = ownPosition.sub(temp);
	if (Math.abs(ownSpeed.x) > 90.) { 
	    if (ownPosition.x > 0.) ownPosition.x -= 105.;
	    else ownPosition.x += 105.;
	    ownSpeed = ownPosition.sub(temp);
	}
    }

    /**
     * calculates own absolute position
     *
     **/
    public void calcAll()
    {
	calcOwnPosition();
	
    }
 
   /**
     * returns own position
     * 
     * @return own Position
     **/
    public CVector getOwnPosition()
    {
	return ownPosition;
    }

    /**
     * returns own speed
     *
     * @return own speed
     **/
    public CVector getOwnSpeed()
    {
	return ownSpeed;
    }

    /**
     * sets ball position in relative polar coordinates
     *
     * @param data string array containing new ball position, DistChange and DirectionChange of the ball
     **/
    public void setBallData( String[] data )
    {
	if ( data == null ) {
	    ballPosition = null;
	    ballChange = null;
	}
	else
	{
	    ballPosition = new PVector( Double.parseDouble( data[1] ), // yes, must be 1, not 0. data[0]="b"
					Integer.parseInt( data[2] ) );
	    if ( data[3] != null) {
		try {
		    ballChange = new PVector( Double.parseDouble( data[3] ),
					  Integer.parseInt( data[4] ) );
		} catch (NumberFormatException nfe) {
		    ballChange = new PVector( Double.parseDouble( data[3] ),
					  (int)Double.parseDouble( data[4] ) );
		}
	    }
	    else ballChange = new PVector(0.,0);
	}
    }

    /**
     * returns ball position
     * 
     * @return ball Position
     **/
    public PVector getBallPosition()
    {
	return ballPosition;
    }

    /**
     * sets the data of a flag
     *
     * @param string array containing flag type and see data
     **/
    public void setFlag( String[] flagStr )
    {
	flag = new Flag( flagStr, playDir );
    }

    /**
     * stores a team member in the list
     * 
     * @param str string array containing relative position and position change
     **/
    public void putTeamMember( String[] data )
    {
	tempTeamMembers.add( new PlayerListElement( data ) );
    }

    /**
     * stores an opponent in the list
     * 
     * @param str string array containing relative position and position change
     **/
    public void putOpponent( String[] data )
    {
	tempOpponents.add( new PlayerListElement( data ) );
    }

    /**
     * returns the list of team members
     *
     * @return teamMembers
     **/
    public LinkedList getTeamMembers()
    {
	return teamMembers;
    }

    /**
     * returns the list of opponents
     *
     * @return opponents
     **/
    public LinkedList getOpponents()
    {
	return opponents;
    }

    /**
     * returns the team member that is next to the opposing goal
     *
     * @return the player next to goal, type is PlayerListElement
     **/
    public PlayerListElement getNextPlayerToGoal()
    {
	Iterator it = teamMembers.iterator();
	double temp, distance = 200.;
	PlayerListElement tempPlayer, nextPlayer = null;
	while (it.hasNext())
	{
	    tempPlayer = (PlayerListElement)it.next();
	    temp = (new CVector(52.5,0)).sub(ownPosition.add(tempPlayer.position.toCVector(direction))).length();
	    if ( temp < distance ) {
		distance = temp;
		nextPlayer = tempPlayer;
	    }
	}

	return nextPlayer;
    }

    /**
     * checks if any team member is nearer to the ball than we are
     *
     * @return true if one is nearer, false if we are nearest to ball in our team
     **/
    public boolean otherPlayerNext2Ball()
    {
	if (ballPosition == null) return true;
	Iterator it = teamMembers.iterator();
	boolean otherIsNearer = false;
	while ( !otherIsNearer && it.hasNext())
	    otherIsNearer = ( ballPosition.diff( ((PlayerListElement)it.next()).position ) < ballPosition.distance );
	return otherIsNearer;
    }

    /**
     * checks if too many opponents are nearer the direction we want to shoot the ball in
     * is not used yet because it has not the wished effect and takes a lot of time
     * @author Annika Niehage, Olga Mirkina
     * @param PVector shootignAim: the direction the player wants to shoot in
     * @return true if one there are more than 3 opponents in the shooting direction, false if the area is free
     **/
    public boolean tooManyOpponents(PVector shootingAim)
	{
	    int counter = 0;
	    Iterator it = opponents.iterator();
	    boolean tooMany = false;
	    while ( (counter < 3) && it.hasNext() )
	    {
		PlayerListElement ple = (PlayerListElement)it.next();
		if ( (ple.position.distance < 10) && ( Math.abs(ple.position.direction - shootingAim.direction) < 20 ) )
		    counter++;
	    }
	    if (counter >= 3) tooMany = true;
	    return tooMany;
	    
	    
	}
    

    /**
     * prints all data for debugging purposes
     **/
    public void printAll()
    {
	System.out.println("playDir: "+playDir+"    Number: "+playerNumber);
	System.out.println("abs, neck body angle: "+direction+"  "+neckAngle+"  "+bodyAngle);
	if (ballPosition != null)
	    System.out.println("ball: ("+ballPosition.distance+","+ballPosition.direction+")");
	if (ballChange != null )
	    System.out.println("ballChange: ("+ballChange.distance+","+ballChange.direction+")");
	System.out.println("ownPosition: ("+ownPosition.x+", "+ownPosition.y+")");
	System.out.println("");
    }




/**
 *calculates ball speed
 *
 * @author Annika Niehage
 **/
    public void calcBallSpeed()
    {

	double prx,pry;
	double erx,ery;
	double vrx, vry;
	CVector absBallPosition = ownPosition.add(ballPosition.toCVector(direction));

	//Formula from the manual
	prx = absBallPosition.x - ownPosition.x;
	pry = absBallPosition.y - ownPosition.y;
	
	if (prx == 0 && pry == 0)
	{
	    ballSpeed.x = 0;
	    ballSpeed.y = 0;
	}

	else
	{
	   
		erx = prx / ballPosition.distance;
		ery = pry / ballPosition.distance;
		
		if (pry == 0)
		    vry = 0;
		else
		    vry = ((Math.PI * ballPosition.distance * ballChange.direction * erx) + (180.0 * ballChange.distance * ery)) / (180.0 * ((ery * ery) + (erx * erx)));
		if (prx == 0)
		    vrx = 0;
		else
		    vrx = (ballChange.distance - (vry * ery)) / erx;
		    
		ballSpeed.x = vrx + ownSpeed.x;
		ballSpeed.y = vry + ownSpeed.y;
	
	}
    }


/**
 * get ball Speed
 *
 * @author Annika Niehage
 * @return CVector ballSpeed: the actual ball speed
 **/
    public CVector getBallSpeed()
    {
	return ballSpeed;
    }

/**
 * calculates object speed
 *
 * @author Annika Niehage
 * @param PlayerListElement player: the object which speed I want to know
 * @return CVector objSpeed: the actual speed of the object
 **/
    public CVector calcObjSpeed(PlayerListElement player)
    {
	//  double tangens;
//  	double tangenssqr;
//  	double tangenssqrt;
//  	CVector objSpeed;

//  	if (player.change!=null) {
//  	    tangens = Math.tan((player.position.direction + direction)*Math.PI/180);
//  	    tangenssqr = Math.pow(tangens,2);
//  	    tangenssqrt = Math.sqrt(1+tangenssqr);
//  	    //speed can be positive and negative
//  	    if(Math.abs(ballPosition.direction + direction) < 90)
//  	    {
//  		objSpeed = new CVector(
//  		    ownSpeed.x + tangenssqrt*player.change.distance - tangens/tangenssqrt*player.position.distance*Math.PI/180 - tangenssqr/tangenssqrt*player.change.distance,
//  	       ownSpeed.y + player.position.distance*player.change.direction*Math.PI/180/tangenssqrt + player.change.distance*tangens/tangenssqrt);
//  	    }
//  	    else
//  	    {
//  		objSpeed = new CVector(
//  		    ownSpeed.x - tangenssqrt*player.change.distance - tangens/tangenssqrt*player.position.distance*Math.PI/180 + tangenssqr/tangenssqrt*player.change.distance,
//  		    ownSpeed.y - player.position.distance*player.change.direction*Math.PI/180/tangenssqrt + player.change.distance*tangens/tangenssqrt);
//  	    }
//  	} else {
//  	    objSpeed = new CVector(0., 0.);
//  	}
//  	return objSpeed;
//      }


	CVector object = ownPosition.add(player.position.toCVector(direction));
	CVector objSpeed = new CVector();
	double prx,pry;
	double erx,ery;
	double vrx, vry;
	
	//Formular from the manual
	prx = object.x - ownPosition.x;
	pry = object.y - ownPosition.y;
	
	
	erx = prx / player.position.distance;
	ery = pry / player.position.distance;
	
	vry = ((Math.PI * player.position.distance * player.change.direction * erx) + (180.0 * player.change.distance * ery)) / (180.0 * ((ery * ery) + (erx * erx)));
	vrx = (player.change.distance - (vry * ery)) / erx;
	
	objSpeed.x = vrx + ownSpeed.x;
	objSpeed.y = vry + ownSpeed.y;
	
	return objSpeed;
    }

} // class Worldmodell


